Require Import machine_int.
Import MachineInt.
Require Import ZArith.
Require Import util.

(********************************************************************)
(* specialized versions of int lemmas (2^32, 2^64, beta, shl, etc.) *)
(********************************************************************)

Open Local Scope Z_scope.

Lemma max_u2Z_32 : forall (a:int 32),
  u2Z a < Zbeta 1.
  intros.
  rewrite Zbeta_power1.
  apply max_u2Z.
Qed.

Lemma max_u2Z_32' : forall (a:int 32),
  u2Z a <= Zbeta 1 - 1.
  intros.
  generalize (max_u2Z_32 a); intro.
  omega.
Qed.

Lemma max_u2Z_64 : forall (a:int 64),
  u2Z a < Zbeta 2.
  intros.
  rewrite <-Zpower_64_Zbeta.
  apply max_u2Z.
Qed.

Lemma zero_extend_u2Z_32 : forall (a:int 32), 
  u2Z (zero_extend 32 a) = u2Z a.
  intros.
  apply zero_extend_u2Z.
Qed.

Lemma add_u2Z_64 : forall (a b:int 64), u2Z a + u2Z b < Zbeta 2 ->
  u2Z (add a b) = u2Z a + u2Z b.
  intros.
  apply add_u2Z.
  rewrite Zpower_64_Zbeta.
  auto.
Qed.

Lemma shr_shrink_u2Z_beta_40 : forall (x:int 40),
  u2Z (shr_shrink 32 x) * Zbeta 1 + u2Z (rem 32 x) = u2Z x.
  intros.
  rewrite Zbeta_power1.
  apply shr_shrink_u2Z.
  omega.
Qed.

Lemma shr_shrink_u2Z_beta_72 : forall (x:int 72),
  u2Z (shr_shrink 32 x) * Zbeta 1 + u2Z (rem 32 x) =
  u2Z x.
  intros.
  rewrite Zbeta_power1.
  apply shr_shrink_u2Z.
  omega.
Qed.

Lemma shl_extend_u2Z_32 : forall (x:int 32),
  u2Z (shl_extend 32 x) = u2Z x * Zbeta 1.
  intros.
  rewrite shl_extend_u2Z.
  rewrite Zbeta_power1.
  reflexivity.
Qed.

Lemma shl_extend_u2Z_64 : forall (x:int 8),
  u2Z (shl_extend 64 x) = u2Z x * Zbeta 2.
  intros.
  rewrite shl_extend_u2Z.
  rewrite Zpower_64_Zbeta.
  reflexivity.
Qed.

Lemma shl_extend_u2Z_32'' : forall (x:int 32),
  u2Z x < Zbeta 1 ->
  u2Z (shl_extend 32 x) + Zbeta 1 <= Zbeta 2.
  intros.
  rewrite <-Zpower_64_Zbeta.
  rewrite Zbeta_power1.
  rewrite Zbeta_power1 in H.
  generalize (shl_extend_u2Z'' 32 H 32); intro.
  rewrite Zpower_is_exp in H0.
  rewrite <-Zbeta_power1 in H0.
  rewrite <-Zbeta_is_exp in H0.
  simpl (1+1) in H0.
  rewrite <-Zbeta_power1.
  rewrite Zpower_64_Zbeta.
  assumption.
Qed.

Lemma concat_u2Z'' : forall (a:int 32) (b:int 32),
  u2Z (a ||| b) = u2Z a * Zbeta 1 + u2Z b.
  intros.
  rewrite Zbeta_power1.
  apply (@concat_u2Z 32 32).
Qed.

Lemma shr_shrink_u2Z_64: forall (x:int 64),
  u2Z x = u2Z (shr_shrink 32 x) * Zbeta 1 + u2Z (rem 32 x).
  intros.
  assert (64 >= 32)%nat.
  omega.
  generalize (shr_shrink_u2Z x H); intro.
  rewrite Zbeta_power1.
  rewrite <-H0.
  auto.
Qed.

Lemma shr_shrink_u2Z'_64: forall (x : int 64),
  u2Z (shr_shrink 32 x) * Zbeta 1 = u2Z x - u2Z (rem 32 x).
  intros.
  rewrite Zbeta_power1.
  rewrite shr_shrink_u2Z'.
  auto.
  auto.
  omega.
Qed.

Lemma rem_u2Z_64: forall (x : int 64),
  u2Z (rem 32 x) = u2Z x - u2Z (shr_shrink 32 x) * Zbeta 1.
  intros.
  rewrite rem_u2Z.
  rewrite Zbeta_power1.
  auto.
  omega.
Qed.

Lemma sign_extend_s2Z_16 : forall (v:int 16),
  s2Z (sign_extend 16 v) = s2Z v.
  intros.
  rewrite sign_extend_s2Z.
  auto.
Qed.

Lemma sign_extend_u2Z_16 : forall (v:int 16),
  0 <= u2Z v < Zpower 2 15 ->
  u2Z (sign_extend 16 v) = u2Z v.
  intros.
  rewrite sign_extend_u2Z.
  auto.
  simpl pred.
  auto.
Qed.

Lemma max_u2Z_umul_32 : forall (a b:int 32),
  u2Z (umul a b) <= (Zbeta 1 - 1) * (Zbeta 1 - 1).
  intros.
  rewrite Zbeta_power1.
  apply max_u2Z_umul.
Qed.

Lemma umul_u2Z_32: forall (a b:int 32), 
  u2Z (umul a b) = u2Z a * u2Z b.
  intros.
  rewrite umul_u2Z.
  reflexivity.
Qed.

