Require Import ZArith.
Require Import List.
Require Import util.
Require Import util_tactic.
Require Import machine_int.
Import MachineInt.
Require Import machine_int_spec.
Require Import sum.
Require Import state.
Require Import mips.
Require Import mips_tactics.
Require Import mips_contrib.
Require Import mapstos.

Open Local Scope asm_heap_scope.
Open Local Scope mips_scope.
Open Local Scope Z_scope.

(*****************************************************************************)

Definition FunAndAddiu P i v iv : assert:=
  fun s s' m h =>
    (P s s' m h) /\ u2Z (gpr.lookup i s) = u2Z (gpr.lookup v s (+) sign_extend_16_32 iv).

Ltac NextAddiu :=
  match goal with 
    | |- {{ ?P }} addiu ?J ?V ?IV ; ?C {{ ?Q }} =>
      let tmp := (eval cbv beta in (FunAndAddiu P J V IV)) in
	apply semax_addiu'' with tmp; [unfold FunAndAddiu | idtac]
  end.

(*****************************************************************************)

Definition k := gpr_s0. (* length of multi-precision integers *)
Definition one := gpr_s1. (* register that contains 1 *)
Definition a := gpr_a0. (* arguments *)
Definition b := gpr_a1.
Definition c := gpr_a2.
Definition t := gpr_t0. (* pointer for sw *)
Definition j := gpr_t1. (* counter  *)
Definition atmp := gpr_t2. (* local variables *)
Definition btmp := gpr_t3.
Definition ctmp := gpr_t4.

Definition multiadd :=
  addiu j gpr_zero zero16;
  addiu t c zero16;
  while_bne j k (
    lwxs atmp j a;
    lwxs btmp j b;
    maddu atmp one;
    maddu btmp one;
    mflhxu ctmp;
    sw ctmp zero16 t;
    addiu t t four16;
    addiu j j one16
  ).

Lemma multiadd_specif : forall nk (Hk:(nk > O)%nat) na nb nc 
  (Hna: 4 * na + 4 * Z_of_nat nk < Zbeta 1) 
  (Hnb: 4 * nb + 4 * Z_of_nat nk < Zbeta 1) 
  (Hnc: 4 * nc + 4 * Z_of_nat nk < Zbeta 1)
  va vb vc (Hva: u2Z va = 4 * na) (Hvb: u2Z vb = 4 * nb) (Hvc: u2Z vc = 4 * nc)
  A B (Ha:length A = nk) (Hb:length B = nk), 
  {{ fun s s' m h => exists C, 
     length C = nk /\
     gpr.lookup one s = one32 /\
     gpr.lookup a s = va /\
     gpr.lookup b s = vb /\
     gpr.lookup c s = vc /\
     u2Z (gpr.lookup k s) = Z_of_nat nk /\
     ((var_e a |--> A) ** (var_e b |--> B) ** (var_e c |--> C)) s s' m h /\
     multiplier.is_null m }}
  multiadd
  {{ fun s s' m h => exists C, length C = nk /\
     ((var_e c |--> C) ** TT) s s' m h /\ 
     Sum nk C + u2Z (multiplier.lo m) * Zbeta nk = Sum nk A + Sum nk B }}.
  intros.
  unfold multiadd.

  (********************* addi j gpr_zero zero16; **********************)

  NextAddiu.

  red; intros.
  inversion_clear H as [C].
  decompose [and] H0; clear H0.
  red.
  split.
  unfold zero16.
  rewrite sign_extend_16_32'; try constructor.
  exists C.
  split; trivial.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split.
  Inde_update_store.
  assumption.
  rewrite gpr.lookup_update'; [idtac |  apply gp_reg_beq_false;  auto].
  elim gpr.lookup_update; try var.
  reflexivity.
  intuition.

  (********************* addi t c zero16; *********************)

  NextAddiu.

  red; intros.
  inversion_clear H.
  inversion_clear H0 as [C].
  decompose [and] H; clear H.
  red.
  split.
  unfold zero16.
  rewrite sign_extend_16_32'; try constructor.
  exists C.
  split; trivial.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split.
  Inde_update_store.
  assumption.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  unfold zero16 in H1; rewrite sign_extend_16_32' in H1; auto.
  constructor.
  rewrite gpr.lookup_update'; [idtac |  apply gp_reg_beq_false;  auto].
  elim gpr.lookup_update; try var.
  reflexivity.

  (********************* while_bne j k *********************)

  apply semax_strengthen_pre with
    (fun s s' m h => exists nj, exists C, 
      length C = nk /\
      gpr.lookup one s = one32 /\
      gpr.lookup a s = va /\
      gpr.lookup b s = vb /\ 
      gpr.lookup c s = vc /\
      u2Z (gpr.lookup k s) = Z_of_nat nk /\
      ((var_e a |--> A) ** (var_e b |--> B) ** (var_e c |--> C)) s s' m h /\
      u2Z (gpr.lookup j s) = Z_of_nat nj /\
      (nj <= nk)%nat /\
      multiplier.utoZ m < Zbeta 1 /\
      u2Z (gpr.lookup t s) = 4 * nc + 4 * Z_of_nat nj /\
      Sum nj C + u2Z (multiplier.lo m) * Zbeta nj = Sum nj A + Sum nj B
    ).

  red; intros.
  inversion_clear H.
  inversion_clear H0.
  inversion_clear H as [C].
  decompose [and] H0; clear H0.
  exists O; exists C.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split.
  rewrite H2.
  Rewrite_u2Z.
  reflexivity.
  Constraint.
  Constraint.
  Constraint.
  split.
  apply le_trans with 1%nat.
  repeat constructor.
  assumption.
  split.
  rewrite multiplier.is_null_utoZ.
  Constraint.
  assumption.
  split.
  rewrite H1.
  rewrite H6.
  Rewrite_u2Z.
  rewrite Hvc.
  simpl Z_of_nat.
  ring.
  Constraint.
  Constraint.
  Constraint.
  simpl.
  rewrite multiplier.is_null_lo.
  Rewrite_u2Z.
  ring.
  Constraint.
  assumption.

  apply semax_weaken_post with
    (fun s s' m h => (exists nj, exists C, 
      length C = nk /\
      gpr.lookup one s = one32 /\
      gpr.lookup a s = va /\
      gpr.lookup b s = vb /\
      gpr.lookup c s = vc /\
      u2Z (gpr.lookup k s) = Z_of_nat nk /\
      ((var_e a |--> A) ** (var_e b |--> B) ** (var_e c |--> C)) s s' m h /\
      u2Z (gpr.lookup j s) = Z_of_nat nj /\
      (nj <= nk)%nat /\
      multiplier.utoZ m < Zbeta 1 /\
      u2Z (gpr.lookup t s) = 4 * nc + 4 * Z_of_nat nj /\
      Sum nj C + u2Z (multiplier.lo m) * Zbeta nj =
      Sum nj A + Sum nj B
    )
    /\
    u2Z (gpr.lookup j s) = u2Z (gpr.lookup k s)
    ).

  red; intros.
  inversion_clear H.
  inversion_clear H0 as [nj].
  inversion_clear H as [C].
  decompose [and] H0; clear H0.
  exists C.
  split; trivial.
  split.
  apply my_sep_trivial with ((var_e a |--> A) ** (var_e b |--> B)).
  assoc_comm H7.
  rewrite H1 in H8; rewrite H6 in H8.
  generalize (Z_of_nat_inj _ _ H8); clear H8; intro; subst nj.
  assumption.

  apply semax_while_bne with (P:=
    fun s s' m h => exists nj, exists C, 
      length C = nk /\
      gpr.lookup one s = one32 /\
      gpr.lookup a s = va /\
      gpr.lookup b s = vb /\
      gpr.lookup c s = vc /\
      u2Z (gpr.lookup k s) = Z_of_nat nk /\
      ((var_e a |--> A) ** (var_e b |--> B) ** 
	(var_e c |--> C)) s s' m h /\
      u2Z (gpr.lookup j s) = Z_of_nat nj /\
      (nj <= nk)%nat /\
      multiplier.utoZ m < Zbeta 1 /\
      u2Z (gpr.lookup t s) = 4*nc + 4*Z_of_nat nj /\
      Sum nj C + u2Z (multiplier.lo m) * Zbeta nj =
      Sum nj A + Sum nj B
  ).

  (********************* lwxs atmp j a; *********************)

  apply semax_lwxs_backwards'' with (
    fun s s' m h => exists nj, exists C, 
      length C = nk /\
      gpr.lookup one s = one32 /\
      gpr.lookup a s = va /\
      gpr.lookup b s = vb /\
      gpr.lookup c s = vc /\
      u2Z (gpr.lookup k s) = Z_of_nat nk /\
      ((var_e a |--> A) ** (var_e b |--> B) ** 
	(var_e c |--> C)) s s' m h /\
      u2Z (gpr.lookup j s) = Z_of_nat nj /\
      (nj < nk)%nat /\
      multiplier.utoZ m < Zbeta 1 /\
      u2Z (gpr.lookup t s) = 4*nc + 4*Z_of_nat nj /\
      Sum nj C + u2Z (multiplier.lo m) * Zbeta nj =
      Sum nj A + Sum nj B /\
      gpr.lookup atmp s = nth nj A zero32
  ).

  red; intros.
  inversion_clear H.
  inversion_clear H0 as [nj].
  inversion_clear H as [C].
  decompose [and] H0; clear H0.
  exists (nth nj A zero32).

  assert (Htmp : u2Z (eval (var_e a) s) + 4 * Z_of_nat nk < Zbeta 1 /\
    eval (add_e (var_e a) (shl2_e (var_e j))) s =
    eval (add_e (var_e a) (shl2_e (int_e (Z2u 32 (Z_of_nat nj))))) s /\
    eval (add_e (add_e (var_e a) (shl2_e (var_e j))) (int_e four32)) s =
    eval (add_e (add_e (var_e a) (shl2_e (var_e j))) (int_e four32)) s).
  simpl eval.
  split.
  rewrite H2.
  Constraint.
  split.
  cutrewrite ( gpr.lookup j s = Z2u 32 (Z_of_nat nj) ).
  reflexivity.
  apply u2Z_injection.
  rewrite H8.
  rewrite Z2u_injection.
  reflexivity.
  generalize (min_u2Z va); intro.
  Constraint.
  reflexivity.

  Decompose_list_32 A nj A1 A2.
  lapply (mapstos_decompose_generic_P _ _ _ _ _ _ 
    (var_e a) (add_e (var_e a) (shl2_e (var_e j))) (add_e (add_e (var_e a) (shl2_e (var_e j))) (int_e four32)) 
    s ((var_e b |--> B) ** (var_e c |--> C)) Ha H14 H0 s' m h).
  intro.
  assert ( (((var_e a |--> A) ** (var_e b |--> B)) ** (var_e c |--> C))
    s s' m h -> ((var_e a |--> A) ** ((var_e b |--> B) ** (var_e c |--> C)))
    s s' m h).
  intro.
  assoc_comm H15.
  generalize (H12 (H15 H7)); clear H15 H12; intro.
  simpl app in H12.
  Monotony H12.
  Adjunction H12.
  
  red.
  exists nj; exists C.
  split; trivial.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  rewrite gpr.lookup_update'; [idtac |  apply gp_reg_beq_false;  auto].
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split.
  Inde_update_store.
  
  assoc_comm H12.
  eapply mapstos_compose_generic with (e':=add_e (var_e a) (shl2_e (var_e j)))
    (e'':=(add_e (add_e (var_e a) (shl2_e (var_e j))) (int_e four32))).
  apply Ha.
  apply H14.
  simpl.
  apply H0.
  exact Htmp.

  assoc_comm H12.
  split; trivial.
  split.
  apply le_neq_lt.
  assumption.
  intro.
  apply H1.
  rewrite H6; rewrite H8; rewrite H15; reflexivity.
  split; trivial.
  split; trivial.
  split; trivial.
  exact Htmp.

  rewrite Ha.
  apply le_neq_lt.
  assumption.
  intro.
  apply H1.
  rewrite H6; rewrite H8; rewrite H0; reflexivity.
  
  (********************* lwxs btmp j b; *********************)

  apply semax_lwxs_backwards'' with (
    fun s s' m h => exists nj, exists C,
      length C = nk /\
      gpr.lookup one s = one32 /\
      gpr.lookup a s = va /\
      gpr.lookup b s = vb /\
      gpr.lookup c s = vc /\
      u2Z (gpr.lookup k s) = Z_of_nat nk /\
      ((var_e a |--> A) ** (var_e b |--> B) ** 
	(var_e c |--> C)) s s' m h /\
      u2Z (gpr.lookup j s) = Z_of_nat nj /\
      (nj < nk)%nat /\
      multiplier.utoZ m < Zbeta 1 /\
      u2Z (gpr.lookup t s) = 4*nc + 4 * Z_of_nat nj /\
      Sum nj C + u2Z (multiplier.lo m) * Zbeta nj = Sum nj A + Sum nj B /\
      gpr.lookup atmp s = nth nj A zero32 /\
      gpr.lookup btmp s = nth nj B zero32
  ).

  red; intros.
  inversion_clear H as [nj].
  inversion_clear H0 as [C].
  decompose [and] H; clear H.
  exists (nth nj B zero32).

  assert (Htmp : u2Z (eval (var_e b) s) + 4 * Z_of_nat nk < Zbeta 1 /\
    eval (add_e (var_e b) (shl2_e (var_e j))) s =
    eval (add_e (var_e b) (shl2_e (int_e (Z2u 32 (Z_of_nat nj))))) s /\
    eval (add_e (add_e (var_e b) (shl2_e (var_e j))) (int_e four32)) s =
    eval (add_e (add_e (var_e b) (shl2_e (var_e j))) (int_e four32)) s).
  simpl eval.
  split.
  rewrite H3.
  Constraint.
  split.
  cutrewrite ( gpr.lookup j s = Z2u 32 (Z_of_nat nj) ).
  reflexivity.
  apply u2Z_injection.
  rewrite H7.
  rewrite Z2u_injection.
  reflexivity.
  generalize (min_u2Z va); intro.
  Constraint.
  reflexivity.

  Decompose_list_32 B nj B1 B2.
  lapply (mapstos_decompose_generic_P _ _ _ _ _ _ 
    (var_e b) (add_e (var_e b) (shl2_e (var_e j))) (add_e (add_e (var_e b) (shl2_e (var_e j))) (int_e four32)) 
    s ((var_e a |--> A) ** (var_e c |--> C)) Hb H14 H s' m h).
  intro.
  assert ( (((var_e a |--> A) ** (var_e b |--> B)) ** (var_e c |--> C))
    s s' m h -> ((var_e b |--> B) ** ((var_e a |--> A) ** (var_e c |--> C)))
    s s' m h).
  intro.
  assoc_comm H15.
  generalize (H12 (H15 H6)); clear H15 H12; intro.
  simpl app in H12.
  Monotony H12.
  Adjunction H12.
  
  red.
  exists nj; exists C.
  split; trivial.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  rewrite gpr.lookup_update'; [idtac |  apply gp_reg_beq_false;  auto].
  elim gpr.lookup_update; try var.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split.
  Inde_update_store.
  
  assoc_comm H12.
  eapply mapstos_compose_generic with (e':=add_e (var_e b) (shl2_e (var_e j)))
    (e'':=(add_e (add_e (var_e b) (shl2_e (var_e j))) (int_e four32))).
  apply Hb.
  apply H14.
  simpl.
  apply H.
  exact Htmp.

  assoc_comm H12.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  exact Htmp.

  rewrite Hb.
  assumption.

  (******************** maddu atmp one; ********************)

  apply semax_maddu'' with (
    fun s s' m h => exists nj, exists C,
      length C = nk /\
      gpr.lookup one s = one32 /\
      gpr.lookup a s = va /\
      gpr.lookup b s = vb /\
      gpr.lookup c s = vc /\
      u2Z (gpr.lookup k s) = Z_of_nat nk /\
      ((var_e a |--> A) ** (var_e b |--> B) ** (var_e c |--> C)) s s' m h /\
      u2Z (gpr.lookup j s) = Z_of_nat nj /\
      (nj < nk)%nat /\
      multiplier.utoZ m < Zpower 2 33 /\
      u2Z (gpr.lookup t s) = 4*nc + 4 * Z_of_nat nj /\ 
      Sum nj C + multiplier.utoZ m * Zbeta nj =
      Sum (S nj) A + Sum nj B /\
      gpr.lookup atmp s = nth nj A zero32 /\
      gpr.lookup btmp s = nth nj B zero32
  ).

  intros.
  inversion_clear H as [nj].
  exists nj.
  inversion_clear H0 as [C].
  decompose [and] H; clear H.
  exists C.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split.
  Inde_mult_maddu.
  split; trivial.
  split; trivial.
  split.
  apply utoZ_maddu_op.
  repeat constructor.
  rewrite <-Zbeta_power1.
  assumption.
  rewrite H2.
  unfold one32.
  rewrite umul_1.
  rewrite zero_extend_u2Z_32.
  apply max_u2Z.
  split; trivial.
  split.
  rewrite multiplier.maddu_utoZ.
  rewrite H2.
  unfold one32.
  rewrite umul_1.
  rewrite H12.
  rewrite zero_extend_u2Z_32.
  rewrite Sum_remove_last.
  rewrite <-Zbeta_power.
  generalize (multiplier.utoZ_lo_beta1 _ H9); intro X; decompose [and] X; clear X.
  rewrite H16.
  fold (zero32).
  apply trans_eq with ( Sum nj C + u2Z (multiplier.lo m) * Zbeta nj
    + Zbeta nj * u2Z (nth nj A zero32)).
  ring.
  rewrite H11.
  ring.
  apply Zlt_trans with (Zbeta 1).
  assumption.
  apply Zlt_le_trans with (Zbeta 2 * (2^^8 - 1)).
  unfold Zbeta.
  simpl.
  Constraint.
  apply Zmult_le_compat_l.
  apply Zminus_le_compat.
  apply Zpower_2_le.
  apply multiplier.acx_size_min.
  Constraint.
  split; assumption.
  
  (********************* maddu btmp one; ********************)

  apply semax_maddu'' with (fun s s' m h => exists nj, exists C,
    length C = nk /\
    gpr.lookup one s = one32 /\
    gpr.lookup a s = va /\
    gpr.lookup b s = vb /\
    gpr.lookup c s = vc /\
    u2Z (gpr.lookup k s) = Z_of_nat nk /\
    ((var_e a |--> A) ** (var_e b |--> B) ** 
      (var_e c |--> C)) s s' m h /\
    u2Z (gpr.lookup j s) = Z_of_nat nj /\
    (nj < nk)%nat /\
    multiplier.utoZ m < Zpower 2 34 /\
    u2Z (gpr.lookup t s) = 4 * nc + 4 * Z_of_nat nj /\
    Sum nj C + multiplier.utoZ m * Zbeta nj = Sum (S nj) A + Sum (S nj) B /\
    gpr.lookup atmp s = nth nj A zero32 /\
    gpr.lookup btmp s = nth nj B zero32
  ).

  intros.
  inversion_clear H as [nj].
  exists nj.
  inversion_clear H0 as [C].
  decompose [and] H; clear H.
  exists C.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split.
  Inde_mult_maddu.
  split; trivial.
  split; trivial.
  split.
  rewrite H2.
  unfold one32.
  rewrite umul_1.
  apply utoZ_maddu_op.
  repeat constructor.
  assumption.
  rewrite zero_extend_u2Z_32.
  apply Zlt_trans with (Zbeta 1).
  apply max_u2Z_32.
  unfold Zbeta.
  simpl.
  Constraint.
  split; trivial.
  split.
  rewrite multiplier.maddu_utoZ.
  rewrite H2.
  unfold one32.
  rewrite umul_1.
  rewrite H14.
  rewrite zero_extend_u2Z_32.
  
  generalize (Sum_remove_last _ B nj); intro.
  rewrite <-Zbeta_power in H.
  unfold zero32.
  rewrite H.
  rewrite Zplus_assoc.
  rewrite <-H11.
  ring.
  apply Zlt_le_trans with (Zpower 2 33). 
  assumption.
  apply Zle_trans with (Zbeta 2 * (2^^8 - 1)).
  unfold Zbeta.
  simpl.
  Constraint.
  apply Zmult_le_compat_l.
  apply Zminus_le_compat.
  apply Zpower_2_le.
  apply multiplier.acx_size_min.
  Constraint.
  split; assumption.

  (******************** mflhxu ctmp; ********************)

  apply semax_mflhxu'' with (fun s s' m h => exists nj, exists C,
    length C = nk /\
    gpr.lookup one s = one32 /\
    gpr.lookup a s = va /\
    gpr.lookup b s = vb /\
    gpr.lookup c s = vc /\
    u2Z (gpr.lookup k s) = Z_of_nat nk /\
    ((var_e a |--> A) ** (var_e b |--> B) ** 
      (var_e c |--> C)) s s' m h /\
    u2Z (gpr.lookup j s) = Z_of_nat nj /\
    (nj < nk)%nat /\
    multiplier.utoZ m < Zbeta 1 /\
    u2Z (gpr.lookup t s) = 4*nc + 4 * Z_of_nat nj /\
    u2Z (gpr.lookup ctmp s) * Zbeta nj + 
    Sum nj C + multiplier.utoZ m * Zbeta (S nj) =
    Sum (S nj) A + Sum (S nj) B /\
    gpr.lookup atmp s = nth nj A zero32 /\
    gpr.lookup btmp s = nth nj B zero32
  ).

  red; intros.
  red.
  inversion_clear H as [nj].
  exists nj.
  inversion_clear H0 as [C].
  exists C.
  decompose [and] H; clear H.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  rewrite gpr.lookup_update'; [idtac |  apply gp_reg_beq_false;  auto].
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split.
  Inde_update_store.
  Inde_mult_mflhxu_op.
  split; trivial.
  split; trivial.
  split.
  apply multiplier.mflhxu_beta2_utoZ.
  apply Zlt_trans with (Zpower 2 34).
  assumption.
  unfold Zbeta.
  simpl.
  Constraint.
  split; trivial.
  split.
  rewrite (multiplier.mflhxu_utoZ m) in H11.
  cutrewrite (Zbeta (S nj) = Zbeta (1 + nj)).
  rewrite Zbeta_is_exp.
  rewrite <-H11.
  ring.
  reflexivity.
  split; assumption.

  (********************* sw ctmp zero16 t; ********************)

  apply semax_sw_backwards'' with (fun s s' m h => exists nj, exists C,
    length C = nk /\
    gpr.lookup one s = one32 /\
    gpr.lookup a s = va /\
    gpr.lookup b s = vb /\
    gpr.lookup c s = vc /\
    u2Z (gpr.lookup k s) = Z_of_nat nk /\
    (((var_e a |--> A) ** (var_e b |--> B)) **
      (var_e c |--> C)) s s' m h /\
    nth nj C zero32 = gpr.lookup ctmp s /\
    u2Z (gpr.lookup j s) = Z_of_nat nj /\
    (nj < nk)%nat /\
    multiplier.utoZ m < Zbeta 1 /\
    u2Z (gpr.lookup t s) = 4*nc + 4 * Z_of_nat nj /\
    Sum (S nj) C + multiplier.utoZ m * Zbeta (S nj) =
    Sum (S nj) A + Sum (S nj) B /\
    gpr.lookup atmp s = nth nj A zero32 /\
    gpr.lookup btmp s = nth nj B zero32
  ).

  red; intros.
  inversion_clear H as [nj].
  inversion_clear H0 as [C].
  decompose [and] H; clear H.
  exists (int_e (nth nj C zero32)).

  assert (Htmp :
    u2Z (eval (var_e c) s) + 4 * Z_of_nat nk < Zbeta 1 /\
    eval (add_e (var_e t) (int_e (sign_extend_16_32 zero16))) s =
    eval (add_e (var_e c) (shl2_e (int_e (Z2u 32 (Z_of_nat nj))))) s /\
    eval (add_e (add_e (var_e t) (int_e (sign_extend_16_32 zero16))) (int_e four32)) s =
    eval (add_e (add_e (var_e t) (int_e (sign_extend_16_32 zero16))) (int_e four32)) s).
  simpl eval.
  rewrite H4.
  rewrite Hvc.
  split; trivial.
  split.
  unfold zero16.
  rewrite sign_extend_16_32'.
  rewrite add_neutral.
  apply u2Z_injection.
  rewrite H10.
  symmetry.
  Rewrite_u2Z.
  rewrite Hvc.
  ring.
  generalize (min_u2Z va); intro.
  Constraint.
  generalize (min_u2Z va); intro.
  Constraint.
  generalize (min_u2Z va); intro.
  Constraint.
  Constraint.
  generalize (min_u2Z va); intro.
  Constraint.
  generalize (min_u2Z va); intro.
  Constraint.
  generalize (min_u2Z va); intro.
  Constraint.
  Constraint.
  reflexivity.

  Decompose_list_32 C nj C1 C2.
  lapply (mapstos_decompose_generic_P _ _ _ _ _ _ 
    (var_e c) (add_e (var_e t) (int_e (sign_extend_16_32 zero16))) 
    (add_e (add_e (var_e t) (int_e (sign_extend_16_32 zero16))) (int_e four32)) s 
    ((var_e b |--> B) ** (var_e a |--> A)) H0 H15 H s' m h).
  intro.
  assert ((((var_e a |--> A) ** (var_e b |--> B)) ** (var_e c |--> C)) s s' m h ->
    ((var_e c |--> C) ** ((var_e b |--> B) ** (var_e a |--> A))) s s' m h).
  intro.
  assoc_comm H16.
  generalize (H13 (H16 H6)); clear H16 H13; intro.
  Monotony H13.
  Adjunction H13.
  exists nj.
  exists (update_list C nj (gpr.lookup ctmp s)).
  split.
  apply length_update_list; auto.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split.

  assoc_comm H13.
  eapply mapstos_compose_generic with (lst2:=C2) (w:=gpr.lookup ctmp s)
    (e':=add_e (var_e t) (int_e (sign_extend_16_32 zero16)))
    (e'':=add_e (add_e (var_e t) (int_e (sign_extend_16_32 zero16))) (int_e four32)).
  apply length_update_list.
  apply H0.
  apply H15.
  rewrite H.
  rewrite update_list_app.
  simpl.
  rewrite H15.
  rewrite <-minus_n_n.
  reflexivity.
  rewrite H15.
  constructor.
  exact Htmp.
  
  simpl app in H13.
  assoc_comm H13.
  generalize H13; apply mapsto_ext'.
  reflexivity.
  simpl.
  reflexivity.
  split.
  rewrite nth_update_list.
  reflexivity.
  rewrite H0.
  assumption.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split.
  rewrite Sum_remove_last.
  rewrite <-Zbeta_power.
  assert ( Sum nj (update_list C nj (gpr.lookup ctmp s)) = Sum nj C ).
  pattern C at 1.
  rewrite H.
  rewrite update_list_app.
  rewrite <-Sum_beyond.
  rewrite H.
  rewrite <-Sum_beyond.
  reflexivity.
  assumption.
  assumption.
  rewrite H15.
  constructor.
  rewrite H16.
  rewrite nth_update_list.
  rewrite <-H11.
  ring.
  rewrite H0.
  assumption.
  split; assumption.

  exact Htmp.
  rewrite H0.
  assumption.

  (********************* addi t t four16; *********************)

  apply semax_addiu'' with (fun s s' m h =>
    exists nj, exists C,
      length C = nk /\
      gpr.lookup one s = one32 /\
      gpr.lookup a s = va /\
      gpr.lookup b s = vb /\
      gpr.lookup c s = vc /\
      u2Z (gpr.lookup k s) = Z_of_nat nk /\
      ((((var_e a |--> A) ** (var_e b |--> B)) ** (var_e c |--> C))) s s' m h /\
      u2Z (gpr.lookup j s) = Z_of_nat nj /\
      (S nj <= nk)%nat /\
      multiplier.utoZ m < Zbeta 1 /\
      u2Z (gpr.lookup t s) = 4*nc + 4 * (Z_of_nat nj + 1) /\
      Sum (S nj) C + u2Z (multiplier.lo m) * Zbeta (S nj) =
      Sum (S nj) A + Sum (S nj) B /\
      gpr.lookup atmp s = nth nj A (zero32) /\
      gpr.lookup btmp s = nth nj B (zero32)).

  red; intros.
  inversion_clear H as [nj].
  inversion_clear H0 as [C].
  decompose [and] H; clear H.
  red.
  exists nj; exists C.
  split; trivial.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  rewrite gpr.lookup_update'; [idtac |  apply gp_reg_beq_false;  auto].
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split.
  Inde_update_store.
  split; trivial.
  split.
  apply lt_le_S.
  assumption.
  split; trivial.
  split.
  Rewrite_u2Z.
  rewrite H11.
  omega.
  Constraint.
  Constraint.
  Constraint.
  split.
  generalize (multiplier.utoZ_lo_beta1 _ H10); intro.
  decompose [and] H; clear H.
  rewrite <-H18.
  assumption.
  split; assumption.

  (********************* addi j j one16 *********************)

  apply semax_addiu'.
  
  red; intros.
  red.
  inversion_clear H as [nj].
  inversion_clear H0 as [C].
  decompose [and] H; clear H.
  exists (S nj); exists C.
  split; trivial.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  rewrite gpr.lookup_update'; [idtac |  apply gp_reg_beq_false;  auto].
  elim gpr.lookup_update; try var.
  elim gpr.lookup_update; try var.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  split.
  Inde_update_store.
  split.
  Rewrite_u2Z.
  OmegaZ.
  Constraint.
  generalize (min_u2Z va); intro.
  Constraint.
  Constraint.
  split; try assumption.
  split; try assumption.
  split; try assumption.
  rewrite Z_S.
  omega.
Qed.


